#include "mvesPolyDataMapper.h"

#include "vesSharedPtr.h"
#include "vesActor.h"
#include "vesBlend.h"
#include "vesDepth.h"
#include "vesGeometryData.h"
#include "vesMapper.h"
#include "vesMaterial.h"
#include "vesRenderer.h"
#include "vesShaderProgram.h"
#include "vesTexture.h"

#include "vesKiwiDataConversionTools.h"

#include <vtkNew.h>
#include <vtkTriangleFilter.h>
#include <vtkLookupTable.h>
#include <vtkExecutive.h>
#include <vtkTrivialProducer.h>

// borrowed from kiwi
static void ConvertVertexArrays(vtkDataSet* dataSet, vesSharedPtr<vesGeometryData> geometryData, vtkScalarsToColors* scalarsToColors=NULL)
{
    vtkUnsignedCharArray* colors = vesKiwiDataConversionTools::FindRGBColorsArray(dataSet);
    vtkDataArray* scalars = vesKiwiDataConversionTools::FindScalarsArray(dataSet);
    vtkDataArray* tcoords = vesKiwiDataConversionTools::FindTextureCoordinatesArray(dataSet);
    if (colors)
    {
        vesKiwiDataConversionTools::SetVertexColors(colors, geometryData);
    }
    else if (scalars)
    {
        vtkSmartPointer<vtkScalarsToColors> colorMap = scalarsToColors;
        if (!colorMap)
        {
            colorMap = vesKiwiDataConversionTools::GetRedToBlueLookupTable(scalars->GetRange());
        }
        vesKiwiDataConversionTools::SetVertexColors(scalars, colorMap, geometryData);
    }
    else if (tcoords)
    {
        vesKiwiDataConversionTools::SetTextureCoordinates(tcoords, geometryData);
    }
}


// borrowed from kiwi
static vesSharedPtr<vesGeometryData> GeometryDataFromPolyData(vtkPolyData* polyData, bool useTriangleFilter) {
    if(polyData->GetNumberOfStrips() == 0 && polyData->GetNumberOfPolys() == 0 && polyData->GetNumberOfLines() == 0) {
            return vesKiwiDataConversionTools::ConvertPoints(polyData);
    }
    if(useTriangleFilter) {
        // Use triangle filter for now to ensure that models containing
        // polygons other than tris and quads will be rendered correctly.
        vtkNew<vtkTriangleFilter> triangleFilter;
        triangleFilter->PassLinesOn();
        triangleFilter->PassVertsOn();
        triangleFilter->SetInputData(polyData);
        triangleFilter->Update();
        
        return vesKiwiDataConversionTools::Convert(triangleFilter->GetOutput());
    } else {
        return vesKiwiDataConversionTools::Convert(polyData);
    }
}


mvesPolyDataMapper::mvesPolyDataMapper() {
    vtkSmartPointer<vtkLookupTable> table = vtkSmartPointer<vtkLookupTable>::New();

    table->SetRange(-1, 1);
    table->SetHueRange(0, 0.666);
    table->Build();
    
    this->setLookupTable(table);
    this->useTriangleFilter = true;
    this->timestamp = vtkSmartPointer<vtkObject>::New();
}

void mvesPolyDataMapper::SetInputConnection(vtkAlgorithmOutput* input) {
    this->input = input;
}

void mvesPolyDataMapper::SetInputData(vtkPolyData* input) {
    vtkSmartPointer<vtkTrivialProducer> p = vtkTrivialProducer::New();
    p->SetOutput(input);
    SetInputConnection(p->GetOutputPort());
}

void mvesPolyDataMapper::setLookupTable(vtkScalarsToColors* lut) {
    this->lookupTable = lut;
}

void mvesPolyDataMapper::setFilterTriangles(bool b) {
    this->useTriangleFilter = b;
}

bool mvesPolyDataMapper::isFilterTriangles() {
    return this->useTriangleFilter;
}

vtkScalarsToColors* mvesPolyDataMapper::getLookupTable() {
    return this->lookupTable;
}

void mvesPolyDataMapper::update() {
    vtkAlgorithm* producer = this->input->GetProducer();
    
    if(producer != NULL) {
        producer->Update();
        vtkPolyData* pd = vtkPolyData::SafeDownCast(producer->GetOutputDataObject(this->input->GetIndex()));
        
        if(pd->GetMTime() > this->timestamp->GetMTime()) {
            vesSharedPtr<vesGeometryData> geometryData = GeometryDataFromPolyData(pd, this->useTriangleFilter);
            ConvertVertexArrays(pd, geometryData, this->lookupTable);
            
            this->setGeometryData(geometryData);
            this->setBoundsDirty(true);
            this->timestamp->Modified();
        }
    }
}

void mvesPolyDataMapper::computeBounds()
{
    update();
    vesMapper::computeBounds();
}

void mvesPolyDataMapper::setColorA(float c[4]) {
    vesMapper::setColor(c[0], c[1], c[2], c[3]);
}

int mvesPolyDataMapper::countCells() {
    vtkAlgorithm* producer = this->input->GetProducer();
    
    if(producer != NULL) {
        producer->Update();
        vtkPolyData* pd = vtkPolyData::SafeDownCast(producer->GetOutputDataObject(this->input->GetIndex()));
        
        return pd->GetNumberOfCells();
    }
    return 0;
}

