#import "mvesView.h"

#include "vesSharedPtr.h"
#include "vesKiwiCameraSpinner.h"
#include "vesKiwiCameraInteractor.h"

#include "mvesRenderer.h"


@interface mvesView() {
@private
    vesSharedPtr<mvesRenderer> renderer;
    vesSharedPtr<vesKiwiCameraSpinner> cameraSpinner;
    vesSharedPtr<vesKiwiCameraInteractor> cameraInteractor;
}

@property (nonatomic, readonly) BOOL motionEnabled;

@end


@implementation mvesView


#pragma mark initializers

- (id)initWithFrame:(CGRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        [self getReady];
    }
    return self;
}

- (void)awakeFromNib
{
    [self getReady];
}


#pragma mark properties

- (void)setBounds:(CGRect)bounds
{
    [super setBounds:bounds];
    self.renderer->resize(self.contentScaleFactor * self.bounds.size.width, self.contentScaleFactor * self.bounds.size.height, 1.0);
}

- (vesSharedPtr<mvesRenderer>)renderer
{
    if(renderer == nil) {
        renderer = vesSharedPtr<mvesRenderer>(new mvesRenderer());
    }
    return renderer;
}

- (void)setMotionEnabled:(BOOL)b
{
    if (b) {
        cameraSpinner->enable();
    }
    else {
        cameraSpinner->disable();
    }
}

- (BOOL)motionEnabled
{
    return cameraSpinner->isEnabled();
}


#pragma mark override UIView methods

- (void)drawRect:(CGRect)rect
{
    self.renderer->resetCameraClippingRange();
    if (self.motionEnabled) {
        cameraSpinner->updateSpin();
    }
    self.renderer->render();
}

- (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{
    self.motionEnabled = NO;
    [super touchesBegan:touches withEvent:event];
}


#pragma mark default gesture recognizers

- (void)addDefaultGestureRecognizers
{
    // one finger pan
    UIPanGestureRecognizer *pgr1 = [[UIPanGestureRecognizer alloc] initWithTarget:self action:@selector(handleOneFingerPan:)];
    pgr1.maximumNumberOfTouches = 1;
    pgr1.minimumNumberOfTouches = 1;
    [self addGestureRecognizer:pgr1];
    
    // two finger pan
    UIPanGestureRecognizer *pgr2 = [[UIPanGestureRecognizer alloc] initWithTarget:self action:@selector(handleTwoFingerPan:)];
    pgr2.maximumNumberOfTouches = 2;
    pgr2.minimumNumberOfTouches = 2;
    [self addGestureRecognizer:pgr2];
    
    // pinch
    [self addGestureRecognizer:[[UIPinchGestureRecognizer alloc] initWithTarget:self action:@selector(handlePinch:)]];
}


#pragma mark camera interaction

- (void)rollCamera:(double)f
{
    cameraInteractor->roll(f);
}

- (void)dollyCamera:(double)f
{
    if (self.renderer->camera()->parallelProjection()) {
        if (f > 0) {
            self.renderer->camera()->setParallelScale(self.renderer->camera()->parallelScale() / f);
        }
    } else {
        cameraInteractor->dolly(f);
    }
}

- (void)panCameraFrom:(CGPoint)p1 to:(CGPoint)p2
{
    cameraInteractor->pan(vesVector2d(p1.x, p1.y), vesVector2d(p2.x, p2.y));
}

- (void)rotateCamera:(CGPoint)v
{
    cameraInteractor->rotate(vesVector2d(v.x, v.y));
}


#pragma mark private API

- (void)getReady
{
    self.context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES2];
    self.drawableDepthFormat = GLKViewDrawableDepthFormat24;
    self.drawableMultisample = GLKViewDrawableMultisample4X;
    [EAGLContext setCurrentContext:self.context];
    
    cameraSpinner = vesSharedPtr<vesKiwiCameraSpinner>(new vesKiwiCameraSpinner());
    cameraInteractor = vesSharedPtr<vesKiwiCameraInteractor>(new vesKiwiCameraInteractor());
    cameraInteractor->setRenderer(self.renderer);
    cameraSpinner->setInteractor(cameraInteractor);
}

- (void)handlePinch:(UIPinchGestureRecognizer *)sender
{
    if (sender.state == UIGestureRecognizerStateChanged || sender.state == UIGestureRecognizerStateEnded) {
        [self dollyCamera:sender.scale];
        sender.scale = 1;
    }
}

- (void)handleOneFingerPan:(UIPanGestureRecognizer *)sender
{
    if (sender.state == UIGestureRecognizerStateChanged) {
        CGFloat s = sender.view.contentScaleFactor;
        CGPoint t = [sender translationInView:sender.view];
        CGPoint p = CGPointMake(t.x * s, t.y * s);
        
        cameraSpinner->handlePanGesture(vesVector2d(p.x, p.y));
        [self rotateCamera:p];
        [sender setTranslation:CGPointZero inView:sender.view];
    }
    if (sender.state == UIGestureRecognizerStateEnded) {
        self.motionEnabled = YES;
    }
}

- (void)handleTwoFingerPan:(UIPanGestureRecognizer *)sender
{
    if (sender.state == UIGestureRecognizerStateChanged) {
        CGFloat s = sender.view.contentScaleFactor;
        CGPoint t = [sender translationInView:sender.view];
        CGPoint p = CGPointMake(t.x * s, -t.y * s);
        
        [self panCameraFrom:CGPointZero to:p];
        [sender setTranslation:CGPointZero inView:sender.view];
    }
}

@end
